from pytorchrl.agent.algorithms.policy_loss_addons.base import PolicyLossAddOn
from pytorchrl.agent.algorithms.policy_loss_addons.kv_similarity import AttractionKL
from pytorchrl.agent.algorithms.policy_loss_addons.kv_similarity import RepulsionKL
from pytorchrl.agent.algorithms.policy_loss_addons.reward_predictor import RewardPredictor
from pytorchrl.agent.algorithms.policy_loss_addons.return_predictor import ReturnPredictor
